import json
from enum import Enum
from azure.identity import DefaultAzureCredential, get_bearer_token_provider
from langchain_core.callbacks import AsyncCallbackManagerForToolRun, CallbackManagerForToolRun
from langchain_core.tools import ToolException
from openai import AsyncAzureOpenAI
from pydantic import BaseModel, Field
from typing import Optional, Type

from foundationallm.langchain.common import FoundationaLLMToolBase
from foundationallm.config import Configuration, UserIdentity
from foundationallm.models.agents import AgentTool
from foundationallm.models.orchestration import ContentArtifact
from foundationallm.models.resource_providers.ai_models import AIModelBase
from foundationallm.models.resource_providers.configuration import APIEndpointConfiguration
from foundationallm.utils import ObjectUtils
from foundationallm.langchain.exceptions import LangChainException
from foundationallm.models.constants import (
    ResourceObjectIdPropertyNames,
    ResourceObjectIdPropertyValues,
    ResourceProviderNames,
    AIModelResourceTypeNames,
    PromptResourceTypeNames
)

class DALLEImageGenerationToolQualityEnum(str, Enum):
    """ Enum for the quality parameter of the DALL-E image generation tool. """
    standard = "standard"
    hd = "hd"

class DALLEImageGenerationToolStyleEnum(str, Enum):
    """ Enum for the style parameter of the DALL-E image generation tool. """
    natural = "natural"
    vivid = "vivid"

class DALLEImageGenerationToolSizeEnum(str, Enum):
    """ Enum for the size parameter of the DALL-E image generation tool. """
    size1024x1024 = "1024x1024"
    size1792x1024 = "1792x1024"
    size1024x1792 = "1024x1792"

class DALLEImageGenerationToolInput(BaseModel):
    """ Input data model for the DALL-E image generation tool. """
    prompt: str = Field(description="Prompt for the DALL-E image generation tool.", example="A cat in the forest.")
    n: int = Field(description="Number of images to generate.", example=1, default=1)
    quality: DALLEImageGenerationToolQualityEnum = Field(description="Quality of the generated images.", default=DALLEImageGenerationToolQualityEnum.hd)
    style: DALLEImageGenerationToolStyleEnum = Field(description="Style of the generated images.", default=DALLEImageGenerationToolStyleEnum.natural)
    size: DALLEImageGenerationToolSizeEnum = Field(description="Size of the generated images.", default=DALLEImageGenerationToolSizeEnum.size1024x1024)

class DALLEImageGenerationTool(FoundationaLLMToolBase):
    """
    DALL-E image generation tool.
    Supports only Azure Identity authentication.
    """
    args_schema: Type[BaseModel] = DALLEImageGenerationToolInput

    def __init__(self, tool_config: AgentTool, objects: dict, user_identity:UserIdentity, config: Configuration):
        """ Initializes the DALLEImageGenerationTool class with the tool configuration,
            exploded objects collection, user identity, and platform configuration. """
        super().__init__(tool_config, objects, user_identity, config)

        ai_model_object_id = self.tool_config.get_resource_object_id_properties(
            ResourceProviderNames.FOUNDATIONALLM_AIMODEL,
            AIModelResourceTypeNames.AI_MODELS,
            ResourceObjectIdPropertyNames.OBJECT_ROLE,
            ResourceObjectIdPropertyValues.MAIN_MODEL
        )
        if ai_model_object_id is None:
            raise LangChainException("The tools's AI models requires a main_model.", 400)

        self.ai_model = ObjectUtils.get_object_by_id(ai_model_object_id.object_id, self.objects, AIModelBase)
        self.api_endpoint = ObjectUtils.get_object_by_id(self.ai_model.endpoint_object_id, self.objects, APIEndpointConfiguration)
        self.client = self._get_client()

    def _run(self,
            prompt: str,
            n: int = 1,
            quality: DALLEImageGenerationToolQualityEnum = DALLEImageGenerationToolQualityEnum.hd,
            style: DALLEImageGenerationToolStyleEnum = DALLEImageGenerationToolStyleEnum.natural,
            size: DALLEImageGenerationToolSizeEnum = DALLEImageGenerationToolSizeEnum.size1024x1024,
            run_manager: Optional[CallbackManagerForToolRun] = None
            ) -> str:
        raise ToolException("This tool does not support synchronous execution. Please use the async version of the tool.")

    async def _arun(self,
            prompt: str,
            n: int = 1,
            quality: DALLEImageGenerationToolQualityEnum = DALLEImageGenerationToolQualityEnum.hd,
            style: DALLEImageGenerationToolStyleEnum = DALLEImageGenerationToolStyleEnum.natural,
            size: DALLEImageGenerationToolSizeEnum = DALLEImageGenerationToolSizeEnum.size1024x1024,
            run_manager: Optional[AsyncCallbackManagerForToolRun] = None
            ) -> str:
        """
        Generate an image using the Azure OpenAI client.
        """
        print(f'Attempting to generate {n} images with a style of {style}, quality of {quality}, and a size of {size}.')
        try:
            result = await self.client.images.generate(
                model = self.ai_model.deployment_name,
                prompt = prompt,
                n = n,
                quality = quality,
                style = style,
                size = size
            )
            content_artifacts = [
                ContentArtifact(
                    id=self.tool_config.name,
                    title=self.tool_config.name,
                    source='tool',
                    filepath=image_data.url,
                    type='image',
                    metadata = {
                        'tool_name': self.tool_config.name,
                        'tool_input': prompt,
                        'revised_tool_input': image_data.revised_prompt
                    }
                )
                for image_data in result.data
                if image_data.revised_prompt and image_data.url
            ]
            return json.loads(result.model_dump_json()), content_artifacts
        except Exception as e:
            print(f'Image generation error code and message: {e.code}; {e}')
            # Specifically handle content policy violation errors.
            if e.code in ['contentFilter', 'content_policy_violation']:
                err = e.message[e.message.find("{"):e.message.rfind("}")+1]
                err_json = err.replace("'", '"')
                err_json = err_json.replace("True", "true").replace("False", "false")
                obj = json.loads(err_json)
                cfr = obj['error']['inner_error']['content_filter_results']
                filtered = [k for k, v in cfr.items() if v['filtered']]
                error_fmt = f"The image generation request resulted in a content policy violation for the following category: {', '.join(filtered)}"
                raise ToolException(error_fmt)
            elif e.code in ['invalidPayload', 'invalid_payload']:
                raise ToolException(f'The image generation request is invalid: {e.message}')
            else:
                raise ToolException(f"An {e.code} error occurred while attempting to generate the requested image: {e.message}")

    def _get_client(self):
        """
        Returns the an AsyncOpenAI client for DALL-E image generation.
        """
        scope = self.api_endpoint.authentication_parameters.get('scope', 'https://cognitiveservices.azure.com/.default')
        # Set up a Azure AD token provider.
        token_provider = get_bearer_token_provider(
            DefaultAzureCredential(exclude_environment_credential=True),
            scope
        )
        return AsyncAzureOpenAI(
            azure_endpoint = self.api_endpoint.url,
            api_version = self.api_endpoint.api_version,
            azure_ad_token_provider = token_provider
        )



from .audio_analysis_service import AudioAnalysisService
from .image_service import ImageService
from .openai_assistants_api_service import OpenAIAssistantsApiService
from .http_client_service import HttpClientService



"""
Encapsulates vector database information.
"""
from typing import Optional
from pydantic import BaseModel
from foundationallm.models.resource_providers.configuration import APIEndpointConfiguration

class VectorDatabaseConfiguration(BaseModel):
    """Vector database metadata model."""
    vector_database : dict
    vector_database_api_endpoint_configuration: APIEndpointConfiguration



from enum import Enum

class CompletionRequestObjectKeys(str, Enum):
    """Enumerator of the Completion Request Object Keys."""
    AZUREAI_AGENT_ID = "AzureAI.AgentService.Agent.Id"
    AZUREAI_AGENT_THREAD_ID = "AzureAI.AgentService.Thread.Id"
    OPENAI_ASSISTANT_ID = "OpenAI.Assistants.Assistant.Id"
    OPENAI_THREAD_ID = "OpenAI.Assistants.Thread.Id"
    GATEWAY_API_ENDPOINT_CONFIGURATION = "GatewayAPIEndpointConfiguration"
    INSTANCE_ID = "FoundationaLLM.InstanceId"
    AGENT_NAME = "Agent.AgentName"
    AGENT_OBJECT_ID = "Agent.AgentObjectId"
    CONTEXT_API_ENDPOINT_CONFIGURATION = "ContextAPIEndpointConfiguration"
    WORKFLOW_INVOCATION_CONVERSATION_FILES = "Workflow.Invocation.ConversationFiles"
    WORKFLOW_INVOCATION_ATTACHED_FILES = "Workflow.Invocation.AttachedFiles"
    TRACING_TRACE_COMPLETION_REQUEST = "Tracing.TraceCompletionRequest"



from typing import Optional

from langchain.agents import AgentExecutor
from langchain_core.callbacks import CallbackManagerForToolRun
from langchain_core.pydantic_v1 import BaseModel, Field
from langchain_core.tools import BaseTool

class BasePandasDataFrameTool(BaseModel):
    """Base tool for interacting with a Pandas DataFrame."""

    #df: DataFrame = Field(exclude = True)
    agent: AgentExecutor = Field(exclude = True)

    # Override BaseTool.Config to appease mypy
    # See https://github.com/pydantic/pydantic/issues/4173
    class Config(BaseTool.Config):
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True
        extra = 'forbid'


class QueryPandasDataFrameTool(BasePandasDataFrameTool, BaseTool):
    """Tool for performing a describe operation on a Pandas DataFrame."""

    name: str = 'query_pandas_dataframe'
    description: str = """
    Input to this tool is a query asking to describe various features of the dataframe.
    Output is the result from the dataframe.
    """

    def _run(
        self,
        query: str,
        run_manager: Optional[CallbackManagerForToolRun] = None
    ) -> str:
        """Execute the query, return the results or an error message."""
        return self.agent.run(query)

# class DescribeQueryBuilderTool(BasePandasDataFrameTool, BaseTool):
#     """Tool for generating a describe query for a Pandas DataFrame."""

#     name: str = 'build_dataframe_describe_query'
#     description: str = """
#     Input to this tool is a feature of the dataframe to describe.
#     Output is a query to describe the particular feature of the dataframe and explain 
#       how results should be returned.
#     """

#     @root_validator(pre=True)
#     def initialize_llm_chain()

#     def _run(

#     )

class TypeConversionTool(BaseTool):
    """Tool for converting data types to the same type so records can be compared."""

    agent: AgentExecutor = Field(exclude=True)

    name: str = 'convert_type'
    description: str = """
    Input to this tool is a value of one type that should be converted to another specified type.
    """

    def _run(
        self,
        query: str,
        run_manager: Optional[CallbackManagerForToolRun] = None
    ) -> str:
        """Execute the query, return the results or an error message."""
        return self.agent.run(query)



"""
Class: FoundationaLLMWorkflowBase
Description: FoundationaLLM base class for tools that uses the agent workflow model for its configuration.
"""
from abc import ABC, abstractmethod
from typing import List, Optional

from azure.identity import DefaultAzureCredential

from foundationallm.config import Configuration, UserIdentity
from foundationallm.models.agents import ExternalAgentWorkflow
from foundationallm.models.messages.message_history_item import MessageHistoryItem
from foundationallm.models.orchestration import CompletionResponse, FileHistoryItem
from foundationallm.telemetry import Telemetry

from .foundationallm_tool_base import FoundationaLLMToolBase

class FoundationaLLMWorkflowBase(ABC):
    """
    FoundationaLLM base class for workflows that uses the agent workflow model for its configuration.
    """
    def __init__(self,
                 workflow_config: ExternalAgentWorkflow,
                 objects: dict,
                 tools: List[FoundationaLLMToolBase],
                 user_identity: UserIdentity,
                 config: Configuration):
        """
        Initializes the FoundationaLLMWorkflowBase class with the workflow configuration.

        Parameters
        ----------
        workflow_config : ExternalAgentWorkflow
            The workflow assigned to the agent.
        objects : dict
            The exploded objects assigned from the agent.
        tools : List[FoundationaLLMToolBase]
            The tools assigned to the agent.
        user_identity : UserIdentity
            The user identity of the user initiating the request.
        config : Configuration
            The application configuration for FoundationaLLM.
        """
        self.workflow_config = workflow_config
        self.objects = objects
        self.tools = tools if tools is not None else []
        self.user_identity = user_identity
        self.config = config
        self.logger = Telemetry.get_logger(self.workflow_config.name)
        self.tracer = Telemetry.get_tracer(self.workflow_config.name)
        self.default_credential = DefaultAzureCredential(exclude_environment_credential=True)

    @abstractmethod
    async def invoke_async(self,
                           operation_id: str,
                           user_prompt: str,
                           user_prompt_rewrite: Optional[str],
                           message_history: List[MessageHistoryItem],
                           file_history: List[FileHistoryItem],
                           conversation_id: Optional[str] = None)-> CompletionResponse:
        """
        Invokes the workflow asynchronously.

        Parameters
        ----------
        operation_id : str
            The unique identifier of the FoundationaLLM operation.
        user_prompt : str
            The user prompt message.
        message_history : List[MessageHistoryItem]
            The message history.
        """
        pass



import boto3
import botocore
import json
from azure.core.credentials import AzureKeyCredential
from azure.identity import DefaultAzureCredential, get_bearer_token_provider
from google.oauth2 import service_account
from langchain_azure_ai.chat_models import AzureAIChatCompletionsModel
from langchain_core.language_models import BaseLanguageModel
from langchain_aws import ChatBedrockConverse
from langchain_google_vertexai import ChatVertexAI
from langchain_openai import AzureChatOpenAI, ChatOpenAI, OpenAI
from openai import AsyncAzureOpenAI as async_aoi
from foundationallm.config import Configuration
from foundationallm.langchain.exceptions import LangChainException
from foundationallm.models.authentication import AuthenticationTypes
from foundationallm.models.language_models import LanguageModelProvider
from foundationallm.models.operations import OperationTypes
from foundationallm.models.resource_providers.ai_models import AIModelBase
from foundationallm.models.resource_providers.configuration import APIEndpointConfiguration
from foundationallm.utils import ObjectUtils

class LanguageModelFactory:

    def __init__(self, objects:dict, config: Configuration):
        self.objects = objects
        self.config = config

    def get_language_model(self,
                           ai_model_object_id:str,
                           override_operation_type: OperationTypes = None,
                           agent_model_parameter_overrides:dict = None,
                          ) -> BaseLanguageModel:
        """
        Create a language model using the specified endpoint settings.

        override_operation_type : OperationTypes - internally override the operation type for the API endpoint.

        Returns
        -------
        BaseLanguageModel
            Returns an API connector for a chat completion model.
        """
        language_model:BaseLanguageModel = None
        api_key = None

        ai_model = ObjectUtils.get_object_by_id(ai_model_object_id, self.objects, AIModelBase)
        if ai_model is None:
            raise LangChainException("AI model configuration settings are missing.", 400)

        api_endpoint = ObjectUtils.get_object_by_id(ai_model.endpoint_object_id, self.objects, APIEndpointConfiguration)
        if api_endpoint is None:
            raise LangChainException("API endpoint configuration settings are missing.", 400)

        match api_endpoint.provider:
            case LanguageModelProvider.AZUREAI:
                if api_endpoint.authentication_type == AuthenticationTypes.AZURE_IDENTITY:
                    try:
                        scope = api_endpoint.authentication_parameters.get('scope', 'https://cognitiveservices.azure.com/.default')
                        credential = DefaultAzureCredential(exclude_environment_credential=True)
                        language_model = AzureAIChatCompletionsModel(
                            endpoint=api_endpoint.url,
                            credential=credential,
                            model_name=ai_model.deployment_name,
                            client_kwargs={
                                "credential_scopes": [scope],
                                "timeout": api_endpoint.timeout_seconds
                            }
                        )
                        if api_endpoint.api_version is not None:
                            language_model.api_version = api_endpoint.api_version

                    except Exception as e:
                        raise LangChainException(f"Failed to create Azure OpenAI API connector: {str(e)}", 500)
                else: # Key-based authentication
                    try:
                        api_key = self.config.get_value(api_endpoint.authentication_parameters.get('api_key_configuration_name'))
                    except Exception as e:
                        raise LangChainException(f"Failed to retrieve API key: {str(e)}", 500)

                    if api_key is None:
                        raise LangChainException("API key is missing from the configuration settings.", 400)
                    credential = AzureKeyCredential(api_key)
                    language_model = AzureAIChatCompletionsModel(
                        endpoint=api_endpoint.url,
                        credential = credential,
                        model_name=ai_model.deployment_name,
                        client_kwargs={
                            "timeout": api_endpoint.timeout_seconds
                        }
                    )
                    if api_endpoint.api_version is not None:
                            language_model.api_version = api_endpoint.api_version
            case LanguageModelProvider.MICROSOFT:
                op_type = api_endpoint.operation_type
                if override_operation_type is not None:
                    op_type = override_operation_type
                if api_endpoint.authentication_type == AuthenticationTypes.AZURE_IDENTITY:
                    try:
                        scope = api_endpoint.authentication_parameters.get('scope', 'https://cognitiveservices.azure.com/.default')
                        # Set up a Azure AD token provider.
                        token_provider = get_bearer_token_provider(
                            DefaultAzureCredential(exclude_environment_credential=True),
                            scope
                        )

                        if op_type == OperationTypes.CHAT:
                            language_model = AzureChatOpenAI(
                                azure_endpoint=api_endpoint.url,
                                api_version=api_endpoint.api_version,
                                openai_api_type='azure_ad',
                                azure_ad_token_provider=token_provider,
                                azure_deployment=ai_model.deployment_name,
                                request_timeout=api_endpoint.timeout_seconds
                            )
                        elif op_type == OperationTypes.ASSISTANTS_API or op_type == OperationTypes.IMAGE_SERVICES:
                            # Assistants API clients can't have deployment as that is assigned at the assistant level.
                            language_model = async_aoi(
                                azure_endpoint=api_endpoint.url,
                                api_version=api_endpoint.api_version,
                                azure_ad_token_provider=token_provider,
                                timeout=api_endpoint.timeout_seconds
                            )
                        else:
                            raise LangChainException(f"Unsupported operation type: {op_type}", 400)

                    except Exception as e:
                        raise LangChainException(f"Failed to create Azure OpenAI API connector: {str(e)}", 500)
                else: # Key-based authentication
                    try:
                        api_key = self.config.get_value(api_endpoint.authentication_parameters.get('api_key_configuration_name'))
                    except Exception as e:
                        raise LangChainException(f"Failed to retrieve API key: {str(e)}", 500)

                    if api_key is None:
                        raise LangChainException("API key is missing from the configuration settings.", 400)

                    if op_type == OperationTypes.CHAT:
                        language_model = AzureChatOpenAI(
                            azure_endpoint=api_endpoint.url,
                            api_key=api_key,
                            api_version=api_endpoint.api_version,
                            azure_deployment=ai_model.deployment_name,
                            request_timeout=api_endpoint.timeout_seconds
                        )
                    elif op_type == OperationTypes.ASSISTANTS_API or op_type == OperationTypes.IMAGE_SERVICES:
                        # Assistants API clients can't have deployment as that is assigned at the assistant level.
                        language_model = async_aoi(
                            azure_endpoint=api_endpoint.url,
                            api_key=api_key,
                            api_version=api_endpoint.api_version,
                            timeout=api_endpoint.timeout_seconds
                        )
                    else:
                        raise LangChainException(f"Unsupported operation type: {op_type}", 400)
            case LanguageModelProvider.OPENAI:
                try:
                    api_key = self.config.get_value(api_endpoint.authentication_parameters.get('api_key_configuration_name'))
                except Exception as e:
                    raise LangChainException(f"Failed to retrieve API key: {str(e)}", 500)

                if api_key is None:
                    raise LangChainException("API key is missing from the configuration settings.", 400)

                language_model = (
                    ChatOpenAI(base_url=api_endpoint.url, api_key=api_key, request_timeout=api_endpoint.timeout_seconds)
                    if api_endpoint.operation_type == OperationTypes.CHAT
                    else OpenAI(base_url=api_endpoint.url, api_key=api_key, request_timeout=api_endpoint.timeout_seconds)
                )
            case LanguageModelProvider.BEDROCK:
                boto3_config = botocore.config.Config(connect_timeout=60, read_timeout=api_endpoint.timeout_seconds)

                if api_endpoint.authentication_type == AuthenticationTypes.AZURE_IDENTITY:
                    # Get Azure scope for federated authentication as well as the AWS role ARN (Amazon Resource Name).
                    try:
                        scope = self.config.get_value(api_endpoint.authentication_parameters.get('scope'))
                    except Exception as e:
                        raise LangChainException(f"Failed to retrieve scope: {str(e)}", 500)

                    if scope is None:
                        raise LangChainException("Scope is missing from the configuration settings.", 400)

                    try:
                        role_arn = self.config.get_value(api_endpoint.authentication_parameters.get('role_arn'))
                    except Exception as e:
                        raise LangChainException(f"Failed to retrieve Role ARN: {str(e)}", 500)

                    if role_arn is None:
                        raise LangChainException("Role ARN is missing from the configuration settings.", 400)

                    # Get Azure token for designated scope.
                    az_creds = DefaultAzureCredential(exclude_environment_credential=True)
                    azure_token = az_creds.get_token(scope)

                    # Get AWS STS credentials using Azure token.
                    sts_client = boto3.client('sts')
                    sts_response = sts_client.assume_role_with_web_identity(
                        RoleArn=role_arn,
                        RoleSessionName='assume-role',
                        WebIdentityToken=azure_token.token
                    )
                    creds = sts_response['Credentials']

                    # parse region from the URL, ex: https://bedrock-runtime.us-east-1.amazonaws.com/
                    region = api_endpoint.url.split('.')[1]
                    language_model = ChatBedrockConverse(
                        model= ai_model.deployment_name,
                        region_name = region,
                        aws_access_key_id = creds["AccessKeyId"],
                        aws_secret_access_key = creds["SecretAccessKey"],
                        aws_session_token= creds["SessionToken"],
                        config=boto3_config
                    )
                else: # Key-based authentication
                    try:
                        access_key = self.config.get_value(api_endpoint.authentication_parameters.get('access_key'))
                    except Exception as e:
                        raise LangChainException(f"Failed to retrieve access key: {str(e)}", 500)

                    if access_key is None:
                        raise LangChainException("Access key is missing from the configuration settings.", 400)

                    try:
                        secret_key = self.config.get_value(api_endpoint.authentication_parameters.get('secret_key'))
                    except Exception as e:
                        raise LangChainException(f"Failed to retrieve secret key: {str(e)}", 500)

                    if secret_key is None:
                        raise LangChainException("Secret key is missing from the configuration settings.", 400)

                    # parse region from the URL, ex: https://bedrock-runtime.us-east-1.amazonaws.com/
                    region = api_endpoint.url.split('.')[1]
                    language_model = ChatBedrockConverse(
                        model= ai_model.deployment_name,
                        region_name = region,
                        aws_access_key_id = access_key,
                        aws_secret_access_key = secret_key,
                        config=boto3_config
                    )
            case LanguageModelProvider.VERTEXAI:
                # Only supports service account authentication via JSON credentials stored in key vault.
                # Uses the authentication parameter: service_account_credentials to get the application configuration key for this value.
                try:
                    service_account_credentials_definition = json.loads(self.config.get_value(api_endpoint.authentication_parameters.get('service_account_credentials')))
                except Exception as e:
                    raise LangChainException(f"Failed to retrieve service account credentials: {str(e)}", 500)

                if not service_account_credentials_definition:
                    raise LangChainException("Service account credentials are missing from the configuration settings.", 400)

                service_account_credentials = service_account.Credentials.from_service_account_info(service_account_credentials_definition)
                language_model = ChatVertexAI(
                    model=ai_model.deployment_name,
                    temperature=0,
                    max_tokens=None,
                    max_retries=6,
                    stop=None,
                    credentials=service_account_credentials
                )

        # Set model parameters.
        for key, value in ai_model.model_parameters.items():
            if hasattr(language_model, key):
                setattr(language_model, key, value)

        # Set agent model overrides.
        if agent_model_parameter_overrides is not None:
            for key, value in agent_model_parameter_overrides.items():
                if hasattr(language_model, key):
                    setattr(language_model, key, value)

        return language_model



"""
Class:GatewayTextEmbeddingResponse
Description:  Class representing the response from the Gateway API service for text embedding.
"""
from typing import List
from pydantic import BaseModel

class GatewayTextEmbeddingResponse(BaseModel):
    """
    Class representing an embedding vector operation result.
    """
    embedding_vector: List[float]
    tokens_count: int
